﻿using Digst.OioIdws.Rest.Common;
using Newtonsoft.Json.Linq;
using System;
using System.Collections.Generic;
using System.IdentityModel.Tokens;
using System.Linq;
using System.Net;
using System.Net.Http;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using System.Xml;

namespace Digst.OioIdws.Rest.Client.AccessToken {
    /// <summary>
    /// <see cref="IAccessTokenService"/>
    /// </summary>
    public class AccessTokenService : IAccessTokenService {
        // It is an internal class. So we just reuse the client settings object instead of creating a new.
        private readonly OioIdwsClientSettings _settings;

        /// <summary>
        /// // It is an internal class and it is assumed that the settings has been verified by the client <see cref="OioIdwsClient"/>
        /// </summary>
        public AccessTokenService(OioIdwsClientSettings settings) {
            if (settings == null) {
                throw new ArgumentNullException(nameof(settings));
            }

            _settings = settings;
        }

        private string CreateSamlToken(XmlElement element) {
            var sb = new StringBuilder();
            var settings = new XmlWriterSettings {
                Indent = false,
                NewLineOnAttributes = false,
                OmitXmlDeclaration = true,
                NewLineHandling = NewLineHandling.None,
                NamespaceHandling = NamespaceHandling.Default
            };

            //..settings
            using (var writer = XmlWriter.Create(sb, settings)) {
                element.WriteTo(writer);
            }
            var xmlString = sb.ToString();
            string result = Convert.ToBase64String(Encoding.UTF8.GetBytes(xmlString));
            Console.WriteLine(xmlString);
            Console.WriteLine("");
            Console.WriteLine("");
            Console.WriteLine("");
            Console.WriteLine(result);

            return result;
        }

        public async Task<AccessToken> GetTokenAsync(GenericXmlSecurityToken securityToken, CancellationToken cancellationToken) {
            if (securityToken == null) {
                throw new ArgumentNullException(nameof(securityToken));
            }

            string samlToken = CreateSamlToken(securityToken.TokenXml);

            var requestHandler = new WebRequestHandler {
                ClientCertificates = { _settings.ClientCertificate },
                ClientCertificateOptions = ClientCertificateOption.Manual
            };

            var formFields = new Dictionary<string, string>
                {
                {"saml-token", samlToken}
            };

            if (_settings.DesiredAccessTokenExpiry.HasValue) {
                formFields["should-expire-in"] =
                    ((int)_settings.DesiredAccessTokenExpiry.Value.TotalSeconds).ToString();
            }
            var formBody = new FormUrlEncodedContent(formFields);
            var read = formBody.ReadAsStringAsync();
            read.Wait();
            var requestbody = read.Result;
            ServicePointManager.SecurityProtocol = SecurityProtocolType.Tls12;
            var client = new HttpClient( new HttpLoggingHandler(requestHandler));

            var requestContent = string.Format("saml-token={0}", samlToken);
            //old = new FormUrlEncodedContent(formFields),

            var content = new StringContent(requestContent, Encoding.UTF8, "application/x-www-form-urlencoded");

            var response = await client.PostAsync(
                _settings.AccessTokenIssuerEndpoint,
                content,
                cancellationToken);

            if (!response.IsSuccessStatusCode) {
                //just find the first valid authenticate header we recognize
                var challenge = response.Headers.WwwAuthenticate
                    .Select(x => new {
                        Type = AccessTokenTypeParser.FromString(x.Scheme),
                        x.Parameter
                    })
                    .FirstOrDefault(x => x.Type.HasValue);

                if (challenge != null) {
                    var parms = HttpHeaderUtils.ParseOAuthSchemeParameter(challenge.Parameter);
                    throw new OioIdwsChallengeException(
                        challenge.Type.Value,
                        parms["error"],
                        parms["error_description"],
                        $@"Got unexpected challenge while issuing access token from '{
                                _settings.AccessTokenIssuerEndpoint
                            }'
({response.StatusCode})': {parms["error"]} - {parms["error_description"]}");
                }

                throw new InvalidOperationException(
                    $@"Got unexpected response while issuing access token from '{_settings.AccessTokenIssuerEndpoint}'
{response.StatusCode}: {await response.Content.ReadAsStringAsync()}");
            }

            var json = await response.Content.ReadAsStringAsync();
            var jsonValue = JObject.Parse(json);

            var accessToken = new Rest.Client.AccessToken.AccessToken {
                Value = (string)jsonValue["access_token"],
                ExpiresIn = TimeSpan.FromSeconds((int)jsonValue["expires_in"]),
                RetrievedAtUtc = DateTime.UtcNow,
                Type = ParseAccessTokenType((string)jsonValue["token_type"])
            };

            return accessToken;
        }

        private AccessTokenType ParseAccessTokenType(string str) {
            var type = AccessTokenTypeParser.FromString(str);

            if (!type.HasValue) {
                throw new InvalidOperationException($"Unknown access token type '{str}'");
            }

            return type.Value;
        }
    }
}